#ifndef AICPU_H
#define AICPU_H

#include "aitest.h"

class Target;

class AICpu:public QObject
{
	Q_OBJECT

	public:
		AICpu(QObject *parent = 0);
		~AICpu();
		bool applyAngle1(double angle, bool correction = true);
		bool applyAngle(double angle, bool correction = true);
		bool checkAngle(double angle);
		void fire(Target *target, bool once = false);
		//void generateDistanceData();
		void generateDistanceData2();
		void generateDistanceData3();
		void checkThread1();
		void checkThread2();
		void calculateVelocities();
		void findJob1();
		void findJob2();
		void findJob3();

		void whosShot();
		//void aimBot();
		void aimBot2();
		void escape();
		double dist(QPointF ship, QPointF target);
		void killUfo();

		static bool Shot;
		static bool Hyper;
	private:
		unsigned char n;

		
		
};
 static int frameCnt = 0;






#endif
